-----------------------------------------------------------------------------------------------
-- Client Lua Script for ActionSetProfiles
-- @author daihenka
-----------------------------------------------------------------------------------------------
require "Apollo"
require "GameLib"
require "AbilityBook"

local ActionSetProfiles = {}

--
local setmetatable, pairs, ipairs = setmetatable, pairs, ipairs
local tostring, Print = tostring, Print
local random = math.random
local strformat = string.format
local tinsert, tremove = table.insert, table.remove
--

local Profile = {}
local strXmlDocFilename = "ActionSetProfiles.xml"
local strXmlDocPath     = strXmlDocFilename
local knRandCeil = 50000


-----------------------------------------------------------------------------------------------
-- Utility Functions
-----------------------------------------------------------------------------------------------
local function GetAddon(strAddonName)
	local info = Apollo.GetAddonInfo(strAddonName)

	if info and info.bRunning == 1 then 
		return Apollo.GetAddon(strAddonName)
	end
end


local function GetCurrentLayout(oAddon)
  -- get the spell tiers
  local tSpellTiers = {}
	local tAllAbilities = AbilityBook.GetAbilitiesList() or {}
	for idx, tBaseAbility in pairs(tAllAbilities) do
		if tBaseAbility.bIsActive then
      tSpellTiers[tBaseAbility.nId] = tBaseAbility.nCurrentTier
		end
	end
  
  local wndContainer = oAddon.wndMain:FindChild("BGFrame:AbilityBuilderMain:BottomContainer:CurrentSetSlots:SlotItemContainer")
  -- get the current limited action set
  local arSpellIds = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 }
  local arTiers    = { 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1 }
	for idx, wndCurr in pairs(wndContainer:GetChildren()) do
		arSpellIds[idx] = wndCurr:FindChild("SlotDisplay"):GetAbilityId()
    arTiers[idx] = tSpellTiers[arSpellIds[idx]] or 1
	end
	arSpellIds[9]  = oAddon.wndGadgetSlot:FindChild("GadgetSlotItem:SlotDisplay"):GetAbilityId()
	arSpellIds[10] = oAddon.wndPathSlot:FindChild("SlotDisplay"):GetAbilityId()
  
  return arSpellIds, arTiers
end

-----------------------------------------------------------------------------------------------
-- Profile Class
-----------------------------------------------------------------------------------------------
function Profile:new(o)
  o = o or {}
  setmetatable(o, self)
  self.__index = self
  
  o.nId = o.nId or self:GenerateId()
  o.strName = o.strName or ""
  o.tAbilities = o.tAbilities or { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 }
  o.tTiers = o.tTiers or { 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1 }
  o.nSpecId = o.nSpecId or AbilityBook.GetCurrentSpec()
 
  return o
end

-- Generate an over-the-top id like a boss!
-- Though I'm probably going to regret it later :(
function Profile:GenerateId()
  local tTime = GameLib.GetServerTime()
  local nRandom = random(1, knRandCeil)
  return strformat("%d%d%d%d%d%d%d", tTime.nYear, tTime.nMonth, tTime.nDay, tTime.nHour, tTime.nMinute, tTime.nSecond, nRandom)
end

function Profile:Set(oAddon)
  if oAddon == nil then
    return
  end
  local wndContainer = oAddon.wndMain:FindChild("BGFrame:AbilityBuilderMain:BottomContainer:CurrentSetSlots:SlotItemContainer")
  
  oAddon.bDirtyBit = true
	for idx, wndCurr in pairs(wndContainer:GetChildren()) do -- Find the window to remove
		if wndCurr:FindChild("SlotDisplay"):GetAbilityId() ~= 0 then
			oAddon:HelperDeleteSlot(wndCurr:FindChild("SlotDisplay"), true)
		end
	end
  
  for idx, wndCurr in pairs(wndContainer:GetChildren()) do
    local nSpellId = self.tAbilities[idx]
    oAddon:HelperAddSlot(nSpellId)
    if nSpellId ~= 0 then
      AbilityBook.UpdateSpellTier(nSpellId, self.tTiers[idx])
    end
	end
  oAddon.wndPathSlot:FindChild("SlotDisplay"):SetAbilityId(self.tAbilities[10])
	oAddon:DrawSpellBook()
end

function Profile:CreateMacro()
	local nMacroId = MacrosLib.CreateMacro()
	MacrosLib.SetMacroData(nMacroId, false, "ASP-Spec" .. self.nSpecId ..  ": " .. self.strName, "IconSprites:Icon_SkillMisc_UI_wr_offgrp", "/actionset " .. self.nId)
	Print("Macro has been created for Action Set Profile: \"" .. self.strName .. "\" (Spec " .. self.nSpecId .. ")")
	MacrosLib:SaveMacros()
end

local nWaitForSpec, nSetTiersForProfile
local function SetAbilityTiers(tAbilities, tTiers)
	for i = 1,8 do
		if tAbilities[i] > 0 then
			AbilityBook.UpdateSpellTier(tAbilities[i], tTiers[i] or 1)
		end
	end
end
local function ClearAllAbilityTiers()
	local tAllAbilities = AbilityBook.GetAbilitiesList() or {}
	for _, tAbility in ipairs(tAllAbilities) do
		if (tAbility.nCurrentTier or 1) > 1 then
			AbilityBook.UpdateSpellTier(tAbility.nId, 1)
		end
	end
end



-- Set and Activate the profile
function Profile:Activate(bIsRetry)
	local nCurrentSpec = AbilityBook.GetCurrentSpec() or 1
	if nCurrentSpec ~= self.nSpecId then
		AbilityBook.SetCurrentSpec(self.nSpecId)
		-- wait for spec change
		nWaitForSpec = self.nId
		return
	end
	nWaitForSpec = nil
	
	-- clear the all spell tiers
	ClearAllAbilityTiers()
	
	-- set the new spell tiers
	SetAbilityTiers(self.tAbilities, self.tTiers)
	
	-- set the action set
	local tResultInfo = ActionSetLib.RequestActionSetChanges(self.tAbilities)
	if tResultInfo.eResult == ActionSetLib.CodeEnumLimitedActionSetResult.RestrictedInPVP then
		Print(Apollo.GetString("AbilityBuilder_BuildsCantBeChangedCombat"))
		return
	elseif tResultInfo.eResult ~= ActionSetLib.CodeEnumLimitedActionSetResult.Ok then
		Print(Apollo.GetString("AbilityBuilder_InvalidSetWarning"))

		if tResultInfo.eResult == ActionSetLib.CodeEnumLimitedActionSetResult.MissingTag then
			for idx, strTag in pairs(tResultInfo.tTags) do
				Print(Apollo.GetString("AbilityBuilder_InvalidSetWarning"))
			end
		end
		return
	end
	
	-- validate the spell tiers are set.
	local tAllAbilities = AbilityBook.GetAbilitiesList() or {}
	for idx, nAbilityId in ipairs(self.tAbilities) do
		local nExpectedTier = self.tTiers[idx]
		for _, tAbility in ipairs(tAllAbilities) do
			if tAbility.nId == nAbilityId then
				local nCurrentTier = (tAbility.nCurrentTier or 1)
--				Print(tAbility.strName .. " - E: " .. nExpectedTier .. " -- A: " .. nCurrentTier)
				if nExpectedTier ~= nCurrentTier then
					-- wait 0.25s then attempt it again
					if not bIsRetry then
						nSetTiersForProfile = self.nId
						Apollo.CreateTimer("ActionSetProfiles_ActivateProfile", 0.25, false)
					else
						Print("Failed to set the spell tiers for profile: " .. self.strName .. " (Spec " .. self.nSpecId .. ")")
					end
					return
				end
			end
		end
	end
	ChatSystemLib.PostOnChannel(ChatSystemLib.ChatChannel_Command, "Profile: " .. self.strName .. " (Spec " .. self.nSpecId .. ") activated")
end

function Profile:GetSpellTier(nSpellId)
	for idx, nId in ipairs(self.tAbilities) do
		if nId == nSpellId then
			return self.tTiers[idx]
		end
	end
	return 1
end


--- Update the profile from the current spellbook UI
function Profile:Update(oAddon)
  if oAddon == nil then
    return
  end
  
  local arSpellIds, arTiers = GetCurrentLayout(oAddon)
  self.tAbilities = arSpellIds
  self.tTiers = arTiers
  return true
end

function Profile:GetSpellTagDistribution()
	local tProfileAbilities = {}
	for _, nId in pairs(self.tAbilities) do tProfileAbilities[nId] = true end
	
	local tSpellTagDist = {}	
	for _, eSpellTag in ipairs({Spell.CodeEnumSpellTag.Assault, Spell.CodeEnumSpellTag.Support, Spell.CodeEnumSpellTag.Utility}) do
		local tAllAbilities = AbilityBook.GetAbilitiesList(eSpellTag) or {}
		for idx, tBaseAbility in pairs(tAllAbilities) do
			if tProfileAbilities[tBaseAbility.nId] then
				tSpellTagDist[eSpellTag] = (tSpellTagDist[eSpellTag] or 0) + 1
			end
		end
	end
  
	return tSpellTagDist
end

function Profile:GetUnspentPoints()
	local nSpentPoints = 0
	for idx = 1,8 do
		if (self.tTiers[idx] or 1) > 1 then
			nSpentPoints = nSpentPoints + (self.tTiers[idx] - 1)
			if self.tTiers[idx] >= 5 then nSpentPoints = nSpentPoints + 4 end
			if self.tTiers[idx] >= 9 then nSpentPoints = nSpentPoints + 4 end
		end
	end
	return GameLib.GetTotalAbilityPoints() - nSpentPoints
end

function Profile:Matches(nSpecId, arSpellIds, arTiers)
  if nSpecId ~= self.nSpecId then
    return
  end
  
  for i = 1, #arSpellIds do
    if arSpellIds[i] ~= self.tAbilities[i] or arTiers[i] ~= self.tTiers[i] then
      --Print(strformat("%s failed at idx %d. Ability expected %d - actual %d.  Tier expected %d - actual %d", self.strName, i, arSpellIds[i], self.tAbilities[i], arTiers[i], self.tTiers[i]))
      return
    end
  end
  return true
end



-----------------------------------------------------------------------------------------------
-- Addon Constructor
-----------------------------------------------------------------------------------------------
function ActionSetProfiles:new(o)
  o = o or {}
  setmetatable(o, self)
  self.__index = self
  
  return o
end

-----------------------------------------------------------------------------------------------
-- Addon Initializer
-----------------------------------------------------------------------------------------------
function ActionSetProfiles:Init()
  Apollo.RegisterAddon(self)
end

-----------------------------------------------------------------------------------------------
-- OnLoad Handler
-----------------------------------------------------------------------------------------------
function ActionSetProfiles:OnLoad()
  -- If a dev recommends against this direct path workaround then the other workaround will be used
  strXmlDocPath = strformat("%s\\%s", tostring(Apollo.GetAssetFolder()), strXmlDocFilename)
  if strXmlDocPath == "\\" .. strXmlDocFilename then
    -- Asset folder wasn't returned... there must be problems.
    -- Just use whatever working path is set in the client and pray!
    
    strXmlDocPath = strXmlDocFilename
  end
  
  -- Register event handlers
  Apollo.RegisterEventHandler("AbilityWindowHasBeenToggled", "OnOpenAbilityWindow",  self)
  Apollo.RegisterEventHandler("AbilityWindowHasBeenClosed",  "OnCloseAbilityWindow", self)
	Apollo.RegisterEventHandler("SpecChanged",                 "OnSpecChanged",        self)
  
  -- Register timer handlers -- timer names are global?
  Apollo.RegisterTimerHandler("ActionSetProfiles_HideNotificationContainerTimer", "OnNotificationContainerHideBtn", self)
  Apollo.RegisterTimerHandler("ActionSetProfiles_ActivateProfile", "OnActivateProfileTimer", self)
  
  -- Register slash command
  Apollo.RegisterSlashCommand("actionset",                   "OnSlashCmd",           self)
  
  self.bAutoOpen = false
  self.tProfiles = {}
	
	self.wndClipboard = Apollo.LoadForm(strXmlDocPath, "ClipboardForm", nil, self)
	self.wndClipboard:Show(false, true)
end

function ActionSetProfiles:OnActivateProfileTimer()
	if nSetTiersForProfile then
		local tProfile = self:GetProfileById(nSetTiersForProfile)
		if tProfile then
			tProfile:Activate(true)
		end
	end
end

-----------------------------------------------------------------------------------------------
-- OnSave Handler
-----------------------------------------------------------------------------------------------
function ActionSetProfiles:OnSave(eLevel)
  if eLevel ~= GameLib.CodeEnumAddonSaveLevel.Character then
    return
  end
  
  local tSaveData = {}
  tSaveData.tProfiles = self.tProfiles
  tSaveData.bAutoOpen = self.bAutoOpen
  
  return tSaveData
end

-----------------------------------------------------------------------------------------------
-- OnRestore Handler
-----------------------------------------------------------------------------------------------
function ActionSetProfiles:OnRestore(eLevel, tSavedData)
  if eLevel ~= GameLib.CodeEnumAddonSaveLevel.Character then
    return
  end
  
  if tSavedData.tProfiles ~= nil then
    for _, tProfile in ipairs(tSavedData.tProfiles) do
      tinsert(self.tProfiles, Profile:new(tProfile))
    end
  end
  if tSavedData.bAutoOpen ~= nil then
    self.bAutoOpen = tSavedData.bAutoOpen
  end
  
end


-----------------------------------------------------------------------------------------------
-- ToggleAbilitiesWindow Handlers - <3 Carbine for the extra events!
-----------------------------------------------------------------------------------------------
function ActionSetProfiles:OnOpenAbilityWindow()
  self:Setup()
end

function ActionSetProfiles:Setup()
  self.addon = self.addon or GetAddon("Abilities")
  if self.addon == nil then
    -- we couldn't find the Abilities addon?!
    return
  end
  
  self:DrawExpandButton()
  if self.bAutoOpen then
    self:OnExpandBtn()
  end
end

function ActionSetProfiles:DrawExpandButton()
  if self.btnExpand == nil or not self.btnExpand:IsValid() then
    self.btnExpand = self:FactoryProduce("ExpandBtn", self.addon.wndMain)
  end
  self.btnExpand:SetAnchorPoints(1,1,1,1)
  self.btnExpand:SetAnchorOffsets(-30,-230,28,-100)
end

function ActionSetProfiles:FactoryProduce(strFormName, wndParent, tObject)
	local wndNew
  
  if wndParent ~= nil then
    wndNew = wndParent:FindChildByUserData(tObject)
  end
  
	if not wndNew then
		wndNew = Apollo.LoadForm(strXmlDocPath, strFormName, wndParent, self)
    wndNew:SetData(tObject)
	end
	return wndNew
end

function ActionSetProfiles:OnCloseAbilityWindow()
  self:TearDown()
end

function ActionSetProfiles:TearDown()
  if self.addon == nil then
    return
  end
  
  -- Just in case something went wrong and we have to manually destroy the window
  if self.wndMain ~= nil and self.wndMain:IsValid() then
    self.wndMain:Destroy()
  end
  if self.btnExpand ~= nil and self.btnExpand:IsValid() then
    self.btnExpand:Destroy()
  end
end


-----------------------------------------------------------------------------------------------
-- Slash Command Handler
-----------------------------------------------------------------------------------------------
function ActionSetProfiles:OnSlashCmd(cmd, arg)
		-- get the profile
		if (tonumber(arg) or 0) == 0 then return end
		
		for _, tProfile in ipairs(self.tProfiles) do
			if tProfile.nId == arg then
				tProfile:Activate()
				return
			end
		end
end

-----------------------------------------------------------------------------------------------
-- Window Event Handlers
-----------------------------------------------------------------------------------------------
function ActionSetProfiles:OnCollapseBtn(wndHandler, wndControl)
  if self.wndMain == nil or not self.wndMain:IsValid() then
    return
  end
  self.wndMain:Show(false)
  self.wndMain:FindChild("NotificationContainer"):Show(false, true)
  self.btnExpand:Show(true)
end

function ActionSetProfiles:OnExpandBtn(wndHandler, wndControl)
  if self.wndMain == nil or not self.wndMain:IsValid() then
    self.wndMain = self:FactoryProduce("ProfilesForm", self.addon.wndMain)
  end
  self.wndMain:SetAnchorPoints(1,0.5,1,1)
  self.wndMain:SetAnchorOffsets(-37, 0, 293, -85)
  self.wndMain:Show(true)
  self.wndMain:FindChild("NotificationContainer"):Show(false, true)
  self.btnExpand:Show(false)
  
  self:SelectCurrentLayout()
  
  self:UpdateListItems()
  self:UpdateControls()
end

function ActionSetProfiles:SelectCurrentLayout()
  self.addon = self.addon or GetAddon("Abilities")
  if self.addon == nil then
    return
  end
  
  local arSpellIds, arTiers = GetCurrentLayout(self.addon)
  local nSpecId = AbilityBook.GetCurrentSpec()
  for idx, tProfile in ipairs(self.tProfiles) do
    if tProfile:Matches(nSpecId, arSpellIds, arTiers) then
      self.nSelectedProfile = tProfile
      return
    end
  end
end

function ActionSetProfiles:HelperShowNotification(strMessage)
  if not self.wndMain or not self.wndMain:IsValid() then
    return
  end
  
  self.wndMain:FindChild("NotificationContainer"):Show(true)
  self.wndMain:FindChild("NotificationContainerText"):SetText(strMessage or "")
  Apollo.CreateTimer("ActionSetProfiles_HideNotificationContainerTimer", 1.5, false)
end


---------------------------------------------------------------------------------------------------
-- Profile Functions
---------------------------------------------------------------------------------------------------
function ActionSetProfiles:GetProfileById(nId)
	for _, tProfile in ipairs(self.tProfiles) do
		if nId == tProfile.nId then
			return tProfile
		end
	end
end


function ActionSetProfiles:OnSpecChanged(newSpecIndex, specError)
  if specError == AbilityBook.CodeEnumSpecError.Ok then
		if nWaitForSpec then
			local tProfile = self:GetProfileById(nWaitForSpec)
			if tProfile then
				tProfile:Activate()
			end
		end
		if self.wndMain ~= nil and self.wndMain:IsValid() then
			self.nSelectedProfile = nil
			self:SelectCurrentLayout()
			
			self:UpdateListItems()
			self:UpdateControls()
		end
	else
		nWaitForSpec = nil
  end
end


function ActionSetProfiles:CreateProfile(strName)
  local tProfile = Profile:new{strName = strName}
  tProfile:Update(self.addon)
  tinsert(self.tProfiles, tProfile)
  self.nSelectedProfile = tProfile
  self:UpdateListItems()
  self:UpdateControls()
end

function ActionSetProfiles:RenameProfile(nId, strName)
	local tProfile = self:GetProfileById(nId)
	if tProfile then
		tProfile.strName = strName
		self:UpdateListItems()
		self:UpdateControls()
  end
end

local kstrSpellTagDist = "<P Font=\"CRB_Button\" Align=\"Center\"><T TextColor=\"UI_BtnTextRedNormal\">%d</T> Assault <T TextColor=\"00ffffff\">--</T> <T TextColor=\"UI_TextHoloTitle\">%d</T> Support <T TextColor=\"00ffffff\">--</T> <T TextColor=\"ffffc757\">%d</T> Utility</P>"
function ActionSetProfiles:UpdateListItems()
  local wndContainer = self.wndMain:FindChild("ListContainer")
  wndContainer:DestroyChildren()
  
  local nSpecId = AbilityBook.GetCurrentSpec()
  self.wndMain:FindChild("SpecText"):SetText(strformat("Profiles for Action Set %d", nSpecId))
  
  for idx, tProfile in ipairs(self.tProfiles) do
    if tProfile.nSpecId == nSpecId then
      local wndCurr = self:FactoryProduce("ProfileListItem", wndContainer, tProfile)
			wndCurr:FindChild("Name"):SetText(tProfile.strName)
			
			self:UpdateListItemDetails(wndCurr)
			
      if self.nSelectedProfile ~= nil and tProfile.nId == self.nSelectedProfile.nId then
        wndCurr:SetCheck(true)
      end
    end
  end
  wndContainer:ArrangeChildrenVert()
end

local ktSpellTagFormats = {
	[Spell.CodeEnumSpellTag.Assault] = { strName = "Assault", cr = "UI_BtnTextRedNormal" },
	[Spell.CodeEnumSpellTag.Support] = { strName = "Support", cr = "UI_TextHoloTitle" },
	[Spell.CodeEnumSpellTag.Utility] = { strName = "Utility", cr = "ffffc757" },
}
function ActionSetProfiles:UpdateListItemDetails(wndCurr)
	if wndCurr == nil or not wndCurr:IsValid() then return end
	local tProfile = wndCurr:GetData()
	local tSpellTagDist = tProfile:GetSpellTagDistribution()
	
	local tSegments = {}
	for eSpellTag, tSpellTagInfo in pairs(ktSpellTagFormats) do
		if (tSpellTagDist[eSpellTag] or 0) > 0 then
			table.insert(tSegments, string.format("<T TextColor=\"%s\">%d</T> %s", tSpellTagInfo.cr, tSpellTagDist[eSpellTag], tSpellTagInfo.strName))
		end
	end
	local strSpellTagDist = string.format("<P Font=\"CRB_Button\" Align=\"Center\">%s</P>", table.concat(tSegments, " <T TextColor=\"00ffffff\">--</T> "))
	
	
	wndCurr:FindChild("SpellTagDist"):SetAML(strSpellTagDist)
	local nUnspentPoints = tProfile:GetUnspentPoints()
	if nUnspentPoints == 0 then
		wndCurr:FindChild("UnspentPoints"):SetAML("")
		wndCurr:FindChild("UnspentPoints"):Show(false)
	else
		wndCurr:FindChild("UnspentPoints"):SetAML(string.format("<P Font=\"CRB_Button\" Align=\"Center\">Unused AP: <T TextColor=\"xkcdYellow\"> %d</T></P>", nUnspentPoints)) -- , GameLib.GetTotalAbilityPoints()))
		wndCurr:FindChild("UnspentPoints"):Show(true)
	end
	wndCurr:FindChild("MouseFill"):SetTooltip("Double click to load the profile\nRight click for more options")
end

function ActionSetProfiles:UpdateControls()
  self.addon = self.addon or GetAddon("Abilities")
  self.wndMain:FindChild("RenameBtn"):Enable(self.nSelectedProfile ~= nil)
  self.wndMain:FindChild("SaveBtn"):Enable(self.nSelectedProfile ~= nil) -- and self.addon.bDirtyBit)
  self.wndMain:FindChild("DeleteBtn"):Enable(self.nSelectedProfile ~= nil)
  self.wndMain:FindChild("AutoOpenBtn"):SetCheck(self.bAutoOpen)
end

---------------------------------------------------------------------------------------------------
-- ProfilesForm Functions
---------------------------------------------------------------------------------------------------

function ActionSetProfiles:OnHintMouseEnter( wndHandler, wndControl, x, y )
end

function ActionSetProfiles:OnHintMouseExit( wndHandler, wndControl, x, y )
end

function ActionSetProfiles:OnNewProfileBtn( wndHandler, wndControl, eMouseButton )
  if self.wndNewProfile == nil or not self.wndNewProfile:IsValid() then
    self.wndNewProfile = self:FactoryProduce("ProfileNewForm", self.addon.wndMain)
  end
  
  self.wndNewProfile:FindChild("ProfileNameEdit"):SetText("")
  self.wndNewProfile:Show(true)
  self.wndNewProfile:FindChild("ProfileNameEdit"):SetFocus()
end

function ActionSetProfiles:OnRenameProfileBtn( wndHandler, wndControl, eMouseButton )
  if self.wndRenameProfile == nil or not self.wndRenameProfile:IsValid() then
    self.wndRenameProfile = self:FactoryProduce("ProfileRenameForm", self.addon.wndMain)
  end
  
  self.wndRenameProfile:SetData(self.nSelectedProfile.nId)
  self.wndRenameProfile:FindChild("ProfileNameEdit"):SetText(self.nSelectedProfile.strName)
  self.wndRenameProfile:Show(true)
  self.wndRenameProfile:FindChild("ProfileNameEdit"):SetFocus()
end

function ActionSetProfiles:OnDeleteProfileBtn( wndHandler, wndControl, eMouseButton )
  if self.nSelectedProfile ~= nil then
    for i, tProfile in ipairs(self.tProfiles) do
      if tProfile.nId == self.nSelectedProfile.nId then
        tremove(self.tProfiles, i)
        self.nSelectedProfile = nil
        self:HelperShowNotification(strformat("%s (Spec %d) has been deleted", tProfile.strName, tProfile.nSpecId))
        break
      end
    end
    self:UpdateListItems()
    self:UpdateControls()
  end
end

function ActionSetProfiles:OnSaveProfileBtn( wndHandler, wndControl, eMouseButton )
  if self.nSelectedProfile ~= nil then
    if self.nSelectedProfile:Update(self.addon) then
      self:HelperShowNotification(strformat("%s (Spec %d) has been updated", self.nSelectedProfile.strName, self.nSelectedProfile.nSpecId))
			self:UpdateListItemDetails(self.wndMain:FindChild("ListContainer"):FindChildByUserData(self.nSelectedProfile))
    end
  end
end

function ActionSetProfiles:OnRenameProfileOk(wndHandler, wndControl, eMouseButton)
	local nId = self.wndRenameProfile:GetData()
	local strName = self.wndRenameProfile:FindChild("ProfileNameEdit"):GetText()
	if nId ~= nil and strName ~= nil then
    self:RenameProfile(nId, strName)
	end	
	self:OnRenameProfileCancel()
end

function ActionSetProfiles:OnRenameProfileCancel(wndHandler, wndControl, eMouseButton)
  if self.wndRenameProfile ~= nil and self.wndRenameProfile:IsValid() then
    self.wndRenameProfile:Destroy()
  end
end

function ActionSetProfiles:OnImportProfileBtn( wndHandler, wndControl, eMouseButton )
  if self.wndImportProfile == nil or not self.wndImportProfile:IsValid() then
    self.wndImportProfile = self:FactoryProduce("ImportProfileForm", self.addon.wndMain)
  end
	
	self.wndImportProfile:FindChild("WebsiteListingText"):SetText("● ws-base.com\n● www.wildstar-arkship.de/actionsetbuilder/en")
  self.wndImportProfile:FindChild("UrlEdit"):SetText("")
	self.wndImportProfile:FindChild("NotificationContainer"):Show(false)
  self.wndImportProfile:Show(true)
  self.wndImportProfile:FindChild("UrlEdit"):SetFocus()
end

function ActionSetProfiles:OnImportProfileOk(wndHandler, wndControl, eMouseButton)
	local strUrl = self.wndImportProfile:FindChild("UrlEdit"):GetText()
	
	local fnImport, strImportName
	
	if strUrl:match("ws%-base%.com%/") then
		if type(self.import_wsbase) == "function" then
			fnImport = self.import_wsbase
			strImportName = "ws-base"
		end
	elseif strUrl:match("wildstar%-arkship%.de%/") then
		if type(self.import_wsasde) == "function" then
			fnImport = self.import_wsasde
			strImportName = "ws-arkship"
		end
	end
	
	if type(fnImport) == "function" then
		local tAbilities, tTiers = fnImport(strUrl)
		if type(tAbilities) == "table" then
			local tProfile = Profile:new{strName = "Imported " .. strImportName .. " profile", tAbilities = tAbilities, tTiers = tTiers}
			tinsert(self.tProfiles, tProfile)
			
			self.nSelectedProfile = tProfile
			self:UpdateListItems()
			self:UpdateControls()
		else
			local strError = (type(tAbilities) == "string" and tAbilities or "An error occurred")
			self:ShowImportErrorNotification(strError)
			return
		end
	else
		self:ShowImportErrorNotification("URL is unsupported")
		return
	end
	
	self:OnImportProfileCancel()
end

function ActionSetProfiles:ShowImportErrorNotification(strError)
	self.wndImportProfile:FindChild("NotificationContainerText"):SetText(strError)
	self.wndImportProfile:FindChild("NotificationContainer"):Show(true)
end

function ActionSetProfiles:OnImportNotificationContainerHideBtn()
	self.wndImportProfile:FindChild("NotificationContainer"):Show(false)
end

function ActionSetProfiles:OnImportProfileCancel(wndHandler, wndControl, eMouseButton)
  if self.wndImportProfile ~= nil and self.wndImportProfile:IsValid() then
    self.wndImportProfile:Destroy()
  end
end

function ActionSetProfiles:OnNewProfileOk(wndHandler, wndControl, eMouseButton)
	local strName = self.wndNewProfile:FindChild("ProfileNameEdit"):GetText()
  self:CreateProfile(strName)
	self:OnNewProfileCancel()
end

function ActionSetProfiles:OnNewProfileCancel(wndHandler, wndControl, eMouseButton)
  if self.wndNewProfile ~= nil and self.wndNewProfile:IsValid() then
    self.wndNewProfile:Destroy()
  end
end

function ActionSetProfiles:OnProfileListItemDoubleClick(wndHandler, wndControl, eMouseButton, nLastRelativeMouseX, nLastRelativeMouseY, bDoubleClick)
	if eMouseButton == 1 then
		self:ShowContextMenu(wndHandler)
		
	elseif bDoubleClick then
		self.nSelectedProfile = wndHandler:GetData()
		self.nSelectedProfile:Set(self.addon)
		self:HelperShowNotification(strformat("Changed ability layout to %s (Spec %d)", self.nSelectedProfile.strName, self.nSelectedProfile.nSpecId))
		
		self:UpdateSelectedListItem()
		self:UpdateControls()
	end
end

function ActionSetProfiles:OnProfileListItemCheck( wndHandler, wndControl, eMouseButton )
  self.nSelectedProfile = wndControl:GetData()
  self:UpdateControls()
end

function ActionSetProfiles:UpdateSelectedListItem()
	if self.nSelectedProfile == nil then return end
	self.wndMain:FindChild("ListContainer"):FindChildByUserData(self.nSelectedProfile):SetCheck(true)
--  for key, wndCurr in pairs(self.wndMain:FindChild("ListContainer"):GetChildren()) do
--    wndCurr:SetCheck(self.nSelectedProfile ~= nil and wndCurr:GetData().nId == self.nSelectedProfile.nId)
--  end
end

function ActionSetProfiles:OnAutoOpenCheck( wndHandler, wndControl, eMouseButton )
  self.bAutoOpen = wndControl:IsChecked()
end

function ActionSetProfiles:OnNotificationContainerHideBtn( wndHandler, wndControl, eMouseButton )
  if not self.wndMain or not self.wndMain:IsValid() then
    return
  end
  
  Apollo.StopTimer("ActionSetProfiles_HideNotificationContainerTimer")
  self.wndMain:FindChild("NotificationContainer"):Show(false)
end

function ActionSetProfiles:ShowContextMenu(wndParent)
	if wndParent == nil or not wndParent:IsValid() then return end
	local tCursor = wndParent:GetMouse()
  local tSCursor = Apollo.GetMouse()
	
	if self.wndContextMenu and self.wndContextMenu:IsValid() then
		self.wndContextMenu:Destroy()
	end
	self.wndContextMenu = Apollo.LoadForm(strXmlDocPath, "ContextMenuProfileForm", "TooltipStratum", self)
	self.wndContextMenu:SetData(wndParent:GetData())
	
	local wndButtonList = self.wndContextMenu:FindChild("ButtonList")
	local tContextBtns = {
		["CreateMacro"]   = "Create Macro",
		["ExportWsAsDe"] 	= "Export Wildstar-Arkship.de Profile",
		["ExportWSBase"] 	= "Export ws-base.com Profile",
		["RenameProfile"] = "Rename Profile",
		["SaveProfile"] 	= "Save Profile",
		["DeleteProfile"] = "Delete Profile",
	}
	
	for eType, strText in pairs(tContextBtns) do
		self:HelperBuildContextMenuButton(wndButtonList, eType, strText)
	end

	
  local nWidth = self.wndContextMenu:GetWidth()
  -- sort and size the context menu
  local nHeight = wndButtonList:ArrangeChildrenVert(0)
  
  -- check the context menu boundaries
  local nLeft, nTop, nRight, nBottom = self:GetContextMenuOffsets(tCursor, tSCursor, nWidth, nHeight)
  self.wndContextMenu:SetAnchorOffsets(nLeft, nTop, nRight, nBottom)
	self.wndMain:FindChild("ListContainer"):Enable(false)
	self.wndContextMenu:ToFront()
end

function ActionSetProfiles:GetContextMenuOffsets(wndCursor, scrCursor, nWidth, nHeight)
  local knOffset = 4
  local knHeightBuffer = 4
  local nScreenWidth, nScreenHeight = Apollo.GetScreenSize()
  local nLeft, nTop = scrCursor.x + knOffset, scrCursor.y + knOffset
  
  local nRight = nLeft + nWidth
  local nBottom = nTop + nHeight + knHeightBuffer
  
  return nLeft, nTop, nRight, nBottom
end

function ActionSetProfiles:HelperBuildContextMenuButton(wndButtonList, eButtonType, strButtonText)
	local wndCurr = Apollo.LoadForm(strXmlDocPath, "ContextBtn", wndButtonList, self)
  wndCurr:SetData(eButtonType)
	wndCurr:FindChild("BtnText"):SetText(strButtonText)
	return wndCurr
end

function ActionSetProfiles:OnContextMenuClosed(wndHandler, wndControl)
	if self.wndContextMenu and self.wndContextMenu:IsValid() then
		self.wndContextMenu:Destroy()
	end
	self.wndMain:FindChild("ListContainer"):Enable(true)
end

function ActionSetProfiles:OnContextMenuBtn(wndHandler, wndControl, eMouseButton)
	self:ProcessContextMenuClick(wndHandler:GetData())
	self:OnContextMenuClosed()
end

function ActionSetProfiles:OnContextMenuBtnMouseEnter( wndHandler, wndControl, x, y )
	wndHandler:GetParent():FindChild("BtnText"):SetTextColor("UI_BtnTextGrayListFlyby")
end

function ActionSetProfiles:OnContextMenuBtnMouseExit( wndHandler, wndControl, x, y )
	wndHandler:GetParent():FindChild("BtnText"):SetTextColor("UI_BtnTextGrayListNormal")
end

function ActionSetProfiles:OnContextMenuBtnMouseDown(wndHandler, wndControl, eMouseButton, nLastRelativeMouseX, nLastRelativeMouseY, bDoubleClick, bStopPropagation )
	self:OnContextMenuBtn(wndHandler:GetParent(), wndHandler:GetParent())
end

function ActionSetProfiles:ProcessContextMenuClick(eButtonType)
	if not self.wndContextMenu or not self.wndContextMenu:IsValid() then
		return
	end
	
	local tProfile = self.wndContextMenu:GetData()
	if eButtonType == "CreateMacro" then
		tProfile:CreateMacro()
	
	elseif eButtonType == "SaveProfile" then
		self.nSelectedProfile = tProfile
		self:OnSaveProfileBtn()
		
	elseif eButtonType == "RenameProfile" then
		self.nSelectedProfile = tProfile
		self:OnRenameProfileBtn()
	
	elseif eButtonType == "DeleteProfile" then
		self.nSelectedProfile = tProfile
		self:OnDeleteProfileBtn()

	elseif eButtonType == "ExportWSBase" then
		local strClipboard
		if type(self.export_wsbase) == "function" then
			strClipboard = self.export_wsbase(tProfile.tAbilities, tProfile.tTiers)
		end
		
		if strClipboard then
			self.wndClipboard:SetText(strClipboard)
			self.wndClipboard:CopyTextToClipboard()
      self:HelperShowNotification("URL has been copied to your clipboard")
		end
	
	elseif eButtonType == "ExportWsAsDe" then
		local strClipboard
		if type(self.export_wsasde) == "function" then
			strClipboard = self.export_wsasde(tProfile.tAbilities, tProfile.tTiers)
		end
		
		if strClipboard then
			self.wndClipboard:SetText(strClipboard)
			self.wndClipboard:CopyTextToClipboard()
      self:HelperShowNotification("URL has been copied to your clipboard")
		end
	end
end  

local aspInst = ActionSetProfiles:new()
aspInst:Init()